/*
								+----------------------------------+
								|                                  |
								|       ***  Vector math  ***      |
								|                                  |
								|   Copyright  -tHE SWINe- 2003   |
								|                                  |
								|             Vector.h             |
								|                                  |
								+----------------------------------+
*/

#ifndef __VECTOR_INCLUDED
#define __VECTOR_INCLUDED

float const Pi = 3.14159265357989323846264338327950f;

template <class T> struct Vector3 {
	T x, y, z;
	//
	inline Vector3() {}
	inline Vector3(T t_x, T t_y, T t_z) { x = t_x; y = t_y; z = t_z; }
	//
	inline T Len() { return (T)sqrt(x * x + y * y + z * z); }
	inline T Len2() { return x * x + y * y + z * z; }
	//
	inline void Normalize()
	{
		T t = Len();
		
		if(t != 0) {
			x /= t;
			y /= t;
			z /= t;
		}
	}

	inline void Normalize(T t_len)
	{
		T t = Len();

		if(t != 0) {
			t /= t_len;
			x /= t;
			y /= t;
			z /= t;
		}
	}
	//
	inline void operator = (Vector3<T> &vec)
	{
		x = vec.x;
		y = vec.y;
		z = vec.z;
	}

	inline Vector3<T> operator + (Vector3<T> &vec)
	{
		return Vector3<T>(x + vec.x, y + vec.y, z + vec.z);
	}

	inline Vector3<T> operator - (Vector3<T> &vec)
	{
		return Vector3<T>(x - vec.x, y - vec.y, z - vec.z);
	}

	inline Vector3<T> operator * (T t)
	{
		return Vector3<T>(x * t, y * t, z * t);
	}

	inline void operator += (Vector3<T> &vec)
	{
		x += vec.x;
		y += vec.y;
		z += vec.z;
	}

	inline void operator -= (Vector3<T> &vec)
	{
		x -= vec.x;
		y -= vec.y;
		z -= vec.z;
	}

	inline void operator *= (T t)
	{
		x *= t;
		y *= t;
		z *= t;
	}

	inline T operator / (Vector3<T> &vec) // dot product
	{
		return x * vec.x + y * vec.y + z * vec.z;
	}

	inline Vector3<T> operator * (Vector3<T> &vec) // cross product
	{
		return Vector3<T>(vec.y * z - vec.z * y, vec.z * x - vec.x * z, vec.x * y - vec.y * x);
	}

	inline void operator *= (Vector3<T> &vec) // cross product
	{
		T tx, ty;

		tx = vec.y * z - vec.z * y;
		ty = vec.z * x - vec.x * z;
		z = vec.x * y - vec.y * x;
		x = tx;
		y = ty;
	}

	void Print()
	{
		printf("Vector > %.2f, %.2f, %.2f\n", x, y, z);
	}
};

typedef Vector3<float> Vector3f;
typedef Vector3<int> Vector3i;
typedef Vector3<short> Vector3s;


template <class T> struct Vector2 {
	T x, y;
	//
	inline Vector2() {}
	inline Vector2(T t_x, T t_y) { x = t_x; y = t_y; }
	//
	inline T Len() { return (T)sqrt(x * x + y * y); }
	inline T Len2() { return x * x + y * y; }
	//
	inline void Normalize()
	{
		T t = Len();
		
		if(t != 0) {
			x /= t;
			y /= t;
		}
	}

	inline void Normalize(T t_len)
	{
		T t = Len();

		if(t != 0) {
			t /= t_len;
			x /= t;
			y /= t;
		}
	}
	//
	inline void operator = (Vector2<T> &vec)
	{
		x = vec.x;
		y = vec.y;
	}

	inline Vector2<T> operator + (Vector2<T> &vec)
	{
		return Vector2<T>(x + vec.x, y + vec.y);
	}

	inline Vector2<T> operator - (Vector2<T> &vec)
	{
		return Vector2<T>(x - vec.x, y - vec.y);
	}

	inline Vector2<T> operator * (T t)
	{
		return Vector2<T>(x * t, y * t);
	}

	inline void operator += (Vector2<T> &vec)
	{
		x += vec.x;
		y += vec.y;
	}

	inline void operator -= (Vector2<T> &vec)
	{
		x -= vec.x;
		y -= vec.y;
	}

	inline void operator *= (T t)
	{
		x *= t;
		y *= t;
	}

	inline T operator / (Vector2<T> &vec) // dot product
	{
		return x * vec.x + y * vec.y;
	}
};

typedef Vector2<float> Vector2f;
typedef Vector2<int> Vector2i;
typedef Vector2<short> Vector2s;


enum {
	Pl_Front,
	Pl_Back,
	Pl_Onplane,
	Pl_Split
};

extern float epsilon;

struct Planef {
	Vector3f v_normal;
	float f_dist;
	//
	/*inline*/ Planef() {}
	/*inline*/ Planef(Vector3f &v_pos, Vector3f &v_norm);
	/*inline*/ Planef(Vector3f &v_u, Vector3f &v_v, Vector3f &v_pos);
	//
	/*inline*/ int Tri_Pos(Vector3f &v_a, Vector3f &v_b, Vector3f &v_c);
	/*inline*/ int Vector_Pos(Vector3f &v_vec);
	/*inline*/ float Vector_Dist(Vector3f &v_vec);
	/*inline*/ float Vector_Dot(Vector3f &v_vec);
	//
	/*inline*/ Vector3f Intersect_Ray(Vector3f &v_org, Vector3f &v_dir);
	/*inline*/ float Intersect_Ray_t(Vector3f &v_org, Vector3f &v_dir);
	//
	/*inline*/ int operator == (Planef &p_plane);
};


struct Matrix4f {
	float f[4][4];
	//
	/*inline*/ void Init();
	/*inline*/ void Translate(Vector3f &v_dir);
	/*inline*/ void RotateX(float f_angle);
	/*inline*/ void RotateY(float f_angle);
	/*inline*/ void RotateZ(float f_angle);
	//
	/*inline*/ Matrix4f operator * (Matrix4f &mat);
	/*inline*/ void operator *= (Matrix4f &mat);
	//
	/*inline*/ void operator = (Matrix4f &mat);
	//
	/*inline*/ void Inverse();
	//
	/*inline*/ Vector3f Transform_Pos(Vector3f &v_vec);
	/*inline*/ Vector3f Transform_Dir(Vector3f &v_vec);
	
	/*inline*/ void Scale(float f_scale);

private:
	/*inline*/ void _Self_Mult(Matrix4f &mat);
};


struct CPolygon {
	Vector3f *m_p_vertex;
	int m_n_vertex_num;
	int m_n_vertex_used;
	//
	Planef m_p_normal;

public:
	CPolygon() { m_n_vertex_num = 0; m_n_vertex_used = 0; m_p_vertex = NULL; }
	~CPolygon();
	//
	void Free();
	//
	int Add_Vertex(Vector3f v_vertex);
	int Add_Vertex(Vector3f *v_vertex, int n_num);
	//
	int n_Vertex_Num() { return m_n_vertex_used; }
	Vector3f *p_Vertex(int n_index);
	//
	Planef p_Normal() { return m_p_normal; }
	int Calc_Normal();
	//
	int Plane_Pos(Planef &p_plane);
	//
	int Cut(Planef &p_plane, int n_desired_half);
	int Split(Planef &p_plane, CPolygon &p_new_poly, int n_desired_half);
	//
	int b_Overlap(CPolygon &poly);
	int b_Contain_Point(Vector3f v_point);
	//
	void Write(FILE *p_file);
	int Read(FILE *p_file);
	//
	void operator = (CPolygon &poly);
	//
	void Reverb();
	//
	int b_IsTiny();

private:
	int GrowBuffer(int n_grow_num);
};

#endif // __VECTOR_INCLUDED